/**
 * transform.js
 *
 * Implement the transform function here.
 *
 * Use context data, aspects, subject(s), and the response from the remote data
 * source to generate samples to send to Refocus.
 *
 * Your transform implementation must return an array of valid samples. The size
 * of the array should equal the number of subjects * the number of aspects.
 *
 * Each sample should have a messageBody no longer than 4096 characters. If the
 * message is longer than that it will be truncated at 4096 by default.
 * If your samples need to include information that would be truncated you
 * should shorten the message yourself before setting it on the samples.
 *
 * DO NOT modify the function signature.
 * DO NOT import or require any other modules.
 * DO NOT modify the module.exports (for testing).
 *
 * Define helper functions in exports.helpers below.
 * Define error handlers in exports.errorHandlers below.
 */

module.exports = {

  /**
   * Use this function signature if you collect data from your connection.url
   * for *all* the subjects in a single request.
   *
   * @param {Object} ctx - The context from the Sample Generator
   * @param {Array} aspects - Array of one or more aspects
   * @param {Array} subjects - Array of one or more subjects
   * @param {http.ServerResponse} res - The response object
   * @returns {Array} an array of samples
   */
  transformBulk(ctx, aspects, subjects, res) {
    const samples = [];

    subjects.forEach((subject) => {
      // Traverse the hierarchy from the response body to find this subject.
      let currentNode = res.body;
      const pathArray = subject.absolutePath.split('.');
      pathArray.forEach((nextNode) => currentNode = currentNode[nextNode]);

      // Create a sample for each aspect.
      aspects.forEach((aspect) => {
        const value = currentNode[aspect.name];
        samples.push({
          name: generateSampleName(subject, aspect),
          value: `${value}`,
        });
      });
    });

    return samples;
  },

  contextDefinition: {
    errorValue: {
      required: false,
      default: '-1',
      description: 'An error sample\'s value, e.g. \'-1\'.',
    },
    baseTrustUrl: {
      required: true,
      description: 'The base url of your Trust1 API endpoint, e.g. \'https://api.status.salesforce.com\'.',
    },
    statusLinkUrl: {
      required: true,
      description: 'The base url of your Trust1 endpoint for a sample\'s related links, e.g. \'http://status.salesforce.com/status/\'.',
    },
    statusMap: {
      required: false,
      default: {
        OK: {
          value: '0'
        },
        INFORMATIONAL_NONCORE: {
          value: '1'
        },
        MAINTENANCE_NONCORE: {
          value: '1'
        },
        MINOR_INCIDENT_NONCORE: {
          value: '2'
        },
        MAJOR_INCIDENT_NONCORE: {
          value: '3'
        },
        INFORMATIONAL_CORE: {
          value: '1',
          messageCode: 'CORE'
        },
        MAINTENANCE_CORE: {
          value: '1',
          messageCode: 'CORE'
        },
        MINOR_INCIDENT_CORE: {
          value: '2',
          messageCode: 'CORE'
        },
        MAJOR_INCIDENT_CORE: {
          value: '3',
          messageCode: 'CORE'
        }
      },
      description: 'An object which maps each Trust1 status enum value to a sample value, messageCode and messageBody.',
    },
  },

  /**
   * Helpers (optional)
   *
   * Define helper functions here if you need to be able to test them directly.
   */
  helpers: {
    generateSampleName(subject, aspect) {
      return `${subject.absolutePath}|${aspect.name}`;
    },
  },
};
